@@ -4,8 +4,11 @@
 #include "mupdf/fitz/system.h"
 #include "mupdf/fitz/context.h"
 #include "mupdf/fitz/buffer.h"
+#include "mupdf/fitz/store.h"
 #include "mupdf/fitz/stream.h"
 
+typedef struct fz_jbig2_globals_s fz_jbig2_globals;
+
 fz_stream *fz_open_copy(fz_stream *chain);
 fz_stream *fz_open_null(fz_stream *chain, int len, int offset);
 fz_stream *fz_open_concat(fz_context *ctx, int max, int pad);
@@ -15,14 +18,16 @@ fz_stream *fz_open_aesd(fz_stream *chain, unsigned char *key, unsigned keylen);
 fz_stream *fz_open_a85d(fz_stream *chain);
 fz_stream *fz_open_ahxd(fz_stream *chain);
 fz_stream *fz_open_rld(fz_stream *chain);
-fz_stream *fz_open_dctd(fz_stream *chain, int color_transform);
-fz_stream *fz_open_resized_dctd(fz_stream *chain, int color_transform, int l2factor);
+fz_stream *fz_open_dctd(fz_stream *chain, int color_transform, int l2factor, fz_stream *jpegtables);
 fz_stream *fz_open_faxd(fz_stream *chain,
 	int k, int end_of_line, int encoded_byte_align,
 	int columns, int rows, int end_of_block, int black_is_1);
 fz_stream *fz_open_flated(fz_stream *chain);
 fz_stream *fz_open_lzwd(fz_stream *chain, int early_change);
 fz_stream *fz_open_predict(fz_stream *chain, int predictor, int columns, int colors, int bpc);
-fz_stream *fz_open_jbig2d(fz_stream *chain, fz_buffer *global);
+fz_stream *fz_open_jbig2d(fz_stream *chain, fz_jbig2_globals *globals);
+
+fz_jbig2_globals *fz_load_jbig2_globals(fz_context *ctx, unsigned char *data, int size);
+void fz_free_jbig2_globals_imp(fz_context *ctx, fz_storable *globals);
 
 #endif
