@@ -14,11 +14,11 @@ fz_hash_table *fz_new_hash_table(fz_context *ctx, int initialsize, int keylen, i
 void fz_empty_hash(fz_context *ctx, fz_hash_table *table);
 void fz_free_hash(fz_context *ctx, fz_hash_table *table);
 
-void *fz_hash_find(fz_context *ctx, fz_hash_table *table, void *key);
-void *fz_hash_insert(fz_context *ctx, fz_hash_table *table, void *key, void *val);
-void *fz_hash_insert_with_pos(fz_context *ctx, fz_hash_table *table, void *key, void *val, unsigned *pos);
-void fz_hash_remove(fz_context *ctx, fz_hash_table *table, void *key);
-void fz_hash_remove_fast(fz_context *ctx, fz_hash_table *table, void *key, unsigned pos);
+void *fz_hash_find(fz_context *ctx, fz_hash_table *table, const void *key);
+void *fz_hash_insert(fz_context *ctx, fz_hash_table *table, const void *key, void *val);
+void *fz_hash_insert_with_pos(fz_context *ctx, fz_hash_table *table, const void *key, void *val, unsigned *pos);
+void fz_hash_remove(fz_context *ctx, fz_hash_table *table, const void *key);
+void fz_hash_remove_fast(fz_context *ctx, fz_hash_table *table, const void *key, unsigned pos);
 
 int fz_hash_len(fz_context *ctx, fz_hash_table *table);
 void *fz_hash_get_key(fz_context *ctx, fz_hash_table *table, int idx);
