@@ -14,6 +14,9 @@
 */
 void fz_write_pnm(fz_context *ctx, fz_pixmap *pixmap, char *filename);
 
+void fz_output_pnm_header(fz_output *out, int w, int h, int n);
+void fz_output_pnm_band(fz_output *out, int w, int h, int n, int band, int bandheight, unsigned char *p);
+
 /*
 	fz_write_pam: Save a pixmap as a pam
 
@@ -21,6 +24,9 @@ void fz_write_pnm(fz_context *ctx, fz_pixmap *pixmap, char *filename);
 */
 void fz_write_pam(fz_context *ctx, fz_pixmap *pixmap, char *filename, int savealpha);
 
+void fz_output_pam_header(fz_output *out, int w, int h, int n, int savealpha);
+void fz_output_pam_band(fz_output *out, int w, int h, int n, int band, int bandheight, unsigned char *sp, int savealpha);
+
 /*
 	fz_write_pbm: Save a bitmap as a pbm
 
