@@ -17,6 +17,13 @@ void pdf_update_text_appearance(pdf_document *doc, pdf_obj *obj, char *eventValu
 void pdf_update_combobox_appearance(pdf_document *doc, pdf_obj *obj);
 void pdf_update_pushbutton_appearance(pdf_document *doc, pdf_obj *obj);
 void pdf_update_text_markup_appearance(pdf_document *doc, pdf_annot *annot, fz_annot_type type);
+void pdf_update_ink_appearance(pdf_document *doc, pdf_annot *annot);
+void pdf_update_text_annot_appearance(pdf_document *doc, pdf_annot *annot);
+/*
+	pdf_update_free_text_annot_appearance: update the appearance stream for a free text
+	annotation, basing it on the annoations rectangle and contents.
+*/
+void pdf_update_free_text_annot_appearance(pdf_document *doc, pdf_annot *annot);
 
 /*
 	pdf_set_annot_appearance: update the appearance of an annotation based
@@ -30,6 +37,6 @@ void pdf_set_annot_appearance(pdf_document *doc, pdf_annot *annot, fz_rect *rect
 */
 void pdf_set_markup_appearance(pdf_document *doc, pdf_annot *annot, float color[3], float alpha, float line_thickness, float line_height);
 
-void pdf_set_ink_appearance(pdf_document *doc, pdf_annot *annot);
+void pdf_set_signature_appearance(pdf_document *doc, pdf_annot *annot, char *name, char *dn, char *date);
 
 #endif
