@@ -1,10 +1,6 @@
 #include "mupdf/fitz.h"
 #include "draw-imp.h"
 
-#define QUANT(x,a) (((int)((x) * (a))) / (a))
-#define HSUBPIX 5.0
-#define VSUBPIX 5.0
-
 #define STACK_SIZE 96
 
 /* Enable the following to attempt to support knockout and/or isolated
@@ -14,6 +10,9 @@
 /* Enable the following to help debug group blending. */
 #undef DUMP_GROUP_BLENDS
 
+/* Enable the following to help debug graphics stack pushes/pops */
+#undef DUMP_STACK_CHANGES
+
 typedef struct fz_draw_device_s fz_draw_device;
 
 enum {
@@ -45,7 +44,7 @@ struct fz_draw_device_s
 	fz_scale_cache *cache_x;
 	fz_scale_cache *cache_y;
 	fz_draw_state *stack;
-	int stack_max;
+	int stack_cap;
 	fz_draw_state init_stack[STACK_SIZE];
 };
 
@@ -77,22 +76,46 @@ static void dump_spaces(int x, const char *s)
 
 #endif
 
+#ifdef DUMP_STACK_CHANGES
+#define STACK_PUSHED(A) stack_change(dev, ">" ## A)
+#define STACK_POPPED(A) stack_change(dev, "<" ## A)
+#define STACK_CONVERT(A) stack_change(dev, A)
+
+static void stack_change(fz_draw_device *dev, char *s)
+{
+	int depth = dev->top;
+	int n;
+
+	if (*s != '<')
+		depth--;
+	n = depth;
+	while (n--)
+		fputc(' ', stderr);
+	fprintf(stderr, "%s (%d)\n", s, depth);
+}
+#else
+#define STACK_PUSHED(A) do {} while (0)
+#define STACK_POPPED(A) do {} while (0)
+#define STACK_CONVERT(A) do {} while (0)
+#endif
+
+
 static void fz_grow_stack(fz_draw_device *dev)
 {
-	int max = dev->stack_max * 2;
+	int max = dev->stack_cap * 2;
 	fz_draw_state *stack;
 
 	if (dev->stack == &dev->init_stack[0])
 	{
 		stack = fz_malloc(dev->ctx, sizeof(*stack) * max);
-		memcpy(stack, dev->stack, sizeof(*stack) * dev->stack_max);
+		memcpy(stack, dev->stack, sizeof(*stack) * dev->stack_cap);
 	}
 	else
 	{
 		stack = fz_resize_array(dev->ctx, dev->stack, max, sizeof(*stack));
 	}
 	dev->stack = stack;
-	dev->stack_max = max;
+	dev->stack_cap = max;
 }
 
 /* 'Push' the stack. Returns a pointer to the current state, with state[1]
@@ -103,7 +126,7 @@ push_stack(fz_draw_device *dev)
 {
 	fz_draw_state *state;
 
-	if (dev->top == dev->stack_max-1)
+	if (dev->top == dev->stack_cap-1)
 		fz_grow_stack(dev);
 	state = &dev->stack[dev->top];
 	dev->top++;
@@ -122,6 +145,7 @@ static void emergency_pop_stack(fz_draw_device *dev, fz_draw_state *state)
 	if (state[1].shape != state[0].shape)
 		fz_drop_pixmap(ctx, state[1].shape);
 	dev->top--;
+	STACK_POPPED("emergency");
 	fz_rethrow(ctx);
 }
 
@@ -138,6 +162,7 @@ fz_knockout_begin(fz_draw_device *dev)
 		return state;
 
 	state = push_stack(dev);
+	STACK_PUSHED("knockout");
 
 	fz_pixmap_bbox(dev->ctx, state->dest, &bbox);
 	fz_intersect_irect(&bbox, &state->scissor);
@@ -164,7 +189,7 @@ fz_knockout_begin(fz_draw_device *dev)
 			fz_clear_pixmap(ctx, dest);
 	}
 
-	if (state->blendmode == 0 && isolated)
+	if ((state->blendmode & FZ_BLEND_MODEMASK) == 0 && isolated)
 	{
 		/* We can render direct to any existing shape plane. If there
 		 * isn't one, we don't need to make one. */
@@ -199,6 +224,7 @@ static void fz_knockout_end(fz_draw_device *dev)
 		return;
 	}
 	state = &dev->stack[--dev->top];
+	STACK_POPPED("knockout");
 	if ((state[0].blendmode & FZ_BLEND_KNOCKOUT) == 0)
 		return;
 
@@ -224,7 +250,11 @@ static void fz_knockout_end(fz_draw_device *dev)
 	else
 		fz_blend_pixmap(state[0].dest, state[1].dest, 255, blendmode, isolated, state[1].shape);
 
-	fz_drop_pixmap(dev->ctx, state[1].dest);
+	/* The following test should not be required, but just occasionally
+	 * errors can cause the stack to get out of sync, and this saves our
+	 * bacon. */
+	if (state[0].dest != state[1].dest)
+		fz_drop_pixmap(dev->ctx, state[1].dest);
 	if (state[0].shape != state[1].shape)
 	{
 		if (state[0].shape)
@@ -256,6 +286,9 @@ fz_draw_fill_path(fz_device *devp, fz_path *path, int even_odd, const fz_matrix
 	if (model == NULL)
 		model = fz_device_gray(dev->ctx);
 
+	if (flatness < 0.001f)
+		flatness = 0.001f;
+
 	fz_reset_gel(dev->gel, &state->scissor);
 	fz_flatten_fill_path(dev->gel, path, ctm, flatness);
 	fz_sort_gel(dev->gel);
@@ -308,6 +341,8 @@ fz_draw_stroke_path(fz_device *devp, fz_path *path, fz_stroke_state *stroke, con
 
 	if (linewidth * expansion < 0.1f)
 		linewidth = 1 / expansion;
+	if (flatness < 0.001f)
+		flatness = 0.001f;
 
 	fz_reset_gel(dev->gel, &state->scissor);
 	if (stroke->dash_len > 0)
@@ -358,11 +393,15 @@ fz_draw_clip_path(fz_device *devp, fz_path *path, const fz_rect *rect, int even_
 	fz_colorspace *model;
 	fz_context *ctx = dev->ctx;
 
+	if (flatness < 0.001f)
+		flatness = 0.001f;
+
 	fz_reset_gel(dev->gel, &state->scissor);
 	fz_flatten_fill_path(dev->gel, path, ctm, flatness);
 	fz_sort_gel(dev->gel);
 
 	state = push_stack(dev);
+	STACK_PUSHED("clip path");
 	model = state->dest->colorspace;
 
 	fz_intersect_irect(fz_bound_gel(dev->gel, &bbox), &state->scissor);
@@ -422,6 +461,8 @@ fz_draw_clip_stroke_path(fz_device *devp, fz_path *path, const fz_rect *rect, fz
 
 	if (linewidth * expansion < 0.1f)
 		linewidth = 1 / expansion;
+	if (flatness < 0.001f)
+		flatness = 0.001f;
 
 	fz_reset_gel(dev->gel, &state->scissor);
 	if (stroke->dash_len > 0)
@@ -431,6 +472,7 @@ fz_draw_clip_stroke_path(fz_device *devp, fz_path *path, const fz_rect *rect, fz
 	fz_sort_gel(dev->gel);
 
 	state = push_stack(dev);
+	STACK_PUSHED("clip stroke");
 	model = state->dest->colorspace;
 
 	fz_intersect_irect(fz_bound_gel(dev->gel, &bbox), &state->scissor);
@@ -467,15 +509,18 @@ fz_draw_clip_stroke_path(fz_device *devp, fz_path *path, const fz_rect *rect, fz
 	}
 }
 
+
 static void
-draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_pixmap *msk,
+draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_glyph *glyph,
 	int xorig, int yorig, const fz_irect *scissor)
 {
-	unsigned char *dp, *mp;
+	unsigned char *dp;
 	fz_irect bbox, bbox2;
 	int x, y, w, h;
+	int skip_x, skip_y;
+	fz_pixmap *msk;
 
-	fz_pixmap_bbox_no_ctx(msk, &bbox);
+	fz_glyph_bbox_no_ctx(glyph, &bbox);
 	fz_translate_irect(&bbox, xorig, yorig);
 	fz_intersect_irect(&bbox, scissor); /* scissor < dst */
 
@@ -487,19 +532,28 @@ draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_pixmap *msk,
 	w = bbox.x1 - bbox.x0;
 	h = bbox.y1 - bbox.y0;
 
-	mp = msk->samples + (unsigned int)((y - msk->y - yorig) * msk->w + (x - msk->x - xorig));
+	skip_x = x - glyph->x - xorig;
+	skip_y = y - glyph->y - yorig;
+
 	dp = dst->samples + (unsigned int)(((y - dst->y) * dst->w + (x - dst->x)) * dst->n);
 
-	assert(msk->n == 1);
-
-	while (h--)
+	msk = glyph->pixmap;
+	if (msk == NULL)
 	{
-		if (dst->colorspace)
-			fz_paint_span_with_color(dp, mp, dst->n, w, colorbv);
-		else
-			fz_paint_span(dp, mp, 1, w, 255);
-		dp += dst->w * dst->n;
-		mp += msk->w;
+		fz_paint_glyph(colorbv, dst, dp, glyph, w, h, skip_x, skip_y);
+	}
+	else
+	{
+		unsigned char *mp = msk->samples + skip_y * msk->w + skip_x;
+		while (h--)
+		{
+			if (dst->colorspace)
+				fz_paint_span_with_color(dp, mp, dst->n, w, colorbv);
+			else
+				fz_paint_span(dp, mp, 1, w, 255);
+			dp += dst->w * dst->n;
+			mp += msk->w;
+		}
 	}
 }
 
@@ -511,12 +565,11 @@ fz_draw_fill_text(fz_device *devp, fz_text *text, const fz_matrix *ctm,
 	unsigned char colorbv[FZ_MAX_COLORS + 1];
 	unsigned char shapebv;
 	float colorfv[FZ_MAX_COLORS];
-	fz_matrix tm, trm, trunc_trm;
-	fz_pixmap *glyph;
-	int i, x, y, gid;
+	fz_matrix tm, trm;
+	fz_glyph *glyph;
+	int i, gid;
 	fz_draw_state *state = &dev->stack[dev->top];
 	fz_colorspace *model = state->dest->colorspace;
-	fz_irect scissor;
 
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
 		state = fz_knockout_begin(dev);
@@ -538,22 +591,14 @@ fz_draw_fill_text(fz_device *devp, fz_text *text, const fz_matrix *ctm,
 		tm.e = text->items[i].x;
 		tm.f = text->items[i].y;
 		fz_concat(&trm, &tm, ctm);
-		x = floorf(trm.e);
-		y = floorf(trm.f);
 
-		trunc_trm = trm;
-		trunc_trm.e = QUANT(trm.e - floorf(trm.e), HSUBPIX);
-		trunc_trm.f = QUANT(trm.f - floorf(trm.f), VSUBPIX);
-
-		scissor.x0 = state->scissor.x0 - x;
-		scissor.y0 = state->scissor.y0 - y;
-		scissor.x1 = state->scissor.x1 - x;
-		scissor.y1 = state->scissor.y1 - y;
-
-		glyph = fz_render_glyph(dev->ctx, text->font, gid, &trunc_trm, model, scissor);
+		glyph = fz_render_glyph(dev->ctx, text->font, gid, &trm, model, &state->scissor);
 		if (glyph)
 		{
-			if (glyph->n == 1)
+			fz_pixmap *pixmap = glyph->pixmap;
+			int x = floorf(trm.e);
+			int y = floorf(trm.f);
+			if (pixmap == NULL || pixmap->n == 1)
 			{
 				draw_glyph(colorbv, state->dest, glyph, x, y, &state->scissor);
 				if (state->shape)
@@ -561,17 +606,19 @@ fz_draw_fill_text(fz_device *devp, fz_text *text, const fz_matrix *ctm,
 			}
 			else
 			{
-				fz_matrix mat = {glyph->w, 0.0, 0.0, glyph->h, x + glyph->x, y + glyph->y};
-				fz_paint_image(state->dest, &state->scissor, state->shape, glyph, &mat, alpha * 255);
+				fz_matrix mat;
+				mat.a = pixmap->w; mat.b = mat.c = 0; mat.d = pixmap->h;
+				mat.e = x + pixmap->x; mat.f = y + pixmap->y;
+				fz_paint_image(state->dest, &state->scissor, state->shape, pixmap, &mat, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES));
 			}
-			fz_drop_pixmap(dev->ctx, glyph);
+			fz_drop_glyph(dev->ctx, glyph);
 		}
 		else
 		{
-			fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &trm);
+			fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &tm);
 			if (path)
 			{
-				fz_draw_fill_path(devp, path, 0, &fz_identity, colorspace, color, alpha);
+				fz_draw_fill_path(devp, path, 0, ctm, colorspace, color, alpha);
 				fz_free_path(dev->ctx, path);
 			}
 			else
@@ -593,12 +640,11 @@ fz_draw_stroke_text(fz_device *devp, fz_text *text, fz_stroke_state *stroke,
 	fz_draw_device *dev = devp->user;
 	unsigned char colorbv[FZ_MAX_COLORS + 1];
 	float colorfv[FZ_MAX_COLORS];
-	fz_matrix tm, trm, trunc_trm;
-	fz_pixmap *glyph;
-	int i, x, y, gid;
+	fz_matrix tm, trm;
+	fz_glyph *glyph;
+	int i, gid;
 	fz_draw_state *state = &dev->stack[dev->top];
 	fz_colorspace *model = state->dest->colorspace;
-	fz_irect scissor;
 
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
 		state = fz_knockout_begin(dev);
@@ -619,32 +665,23 @@ fz_draw_stroke_text(fz_device *devp, fz_text *text, fz_stroke_state *stroke,
 		tm.e = text->items[i].x;
 		tm.f = text->items[i].y;
 		fz_concat(&trm, &tm, ctm);
-		x = floorf(trm.e);
-		y = floorf(trm.f);
 
-		trunc_trm = trm;
-		trunc_trm.e = QUANT(trm.e - floorf(trm.e), HSUBPIX);
-		trunc_trm.f = QUANT(trm.f - floorf(trm.f), VSUBPIX);
-
-		scissor.x0 = state->scissor.x0 - x;
-		scissor.y0 = state->scissor.y0 - y;
-		scissor.x1 = state->scissor.x1 - x;
-		scissor.y1 = state->scissor.y1 - y;
-
-		glyph = fz_render_stroked_glyph(dev->ctx, text->font, gid, &trunc_trm, ctm, stroke, scissor);
+		glyph = fz_render_stroked_glyph(dev->ctx, text->font, gid, &trm, ctm, stroke, &state->scissor);
 		if (glyph)
 		{
+			int x = (int)trm.e;
+			int y = (int)trm.f;
 			draw_glyph(colorbv, state->dest, glyph, x, y, &state->scissor);
 			if (state->shape)
 				draw_glyph(colorbv, state->shape, glyph, x, y, &state->scissor);
-			fz_drop_pixmap(dev->ctx, glyph);
+			fz_drop_glyph(dev->ctx, glyph);
 		}
 		else
 		{
-			fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &trm);
+			fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &tm);
 			if (path)
 			{
-				fz_draw_stroke_path(devp, path, stroke, &fz_identity, colorspace, color, alpha);
+				fz_draw_stroke_path(devp, path, stroke, ctm, colorspace, color, alpha);
 				fz_free_path(dev->ctx, path);
 			}
 			else
@@ -665,9 +702,9 @@ fz_draw_clip_text(fz_device *devp, fz_text *text, const fz_matrix *ctm, int accu
 	fz_context *ctx = dev->ctx;
 	fz_irect bbox;
 	fz_pixmap *mask, *dest, *shape;
-	fz_matrix tm, trm, trunc_trm;
-	fz_pixmap *glyph;
-	int i, x, y, gid;
+	fz_matrix tm, trm;
+	fz_glyph *glyph;
+	int i, gid;
 	fz_draw_state *state;
 	fz_colorspace *model;
 
@@ -676,6 +713,7 @@ fz_draw_clip_text(fz_device *devp, fz_text *text, const fz_matrix *ctm, int accu
 	/* If accumulate == 2 then this text object is a continuation */
 
 	state = push_stack(dev);
+	STACK_PUSHED("clip text");
 	model = state->dest->colorspace;
 
 	if (accumulate == 0)
@@ -721,6 +759,7 @@ fz_draw_clip_text(fz_device *devp, fz_text *text, const fz_matrix *ctm, int accu
 		{
 			mask = state->mask;
 			dev->top--;
+			STACK_POPPED("clip text");
 		}
 
 		if (!fz_is_empty_irect(&bbox) && mask)
@@ -729,8 +768,6 @@ fz_draw_clip_text(fz_device *devp, fz_text *text, const fz_matrix *ctm, int accu
 
 			for (i = 0; i < text->len; i++)
 			{
-				fz_irect scissor;
-
 				gid = text->items[i].gid;
 				if (gid < 0)
 					continue;
@@ -738,45 +775,36 @@ fz_draw_clip_text(fz_device *devp, fz_text *text, const fz_matrix *ctm, int accu
 				tm.e = text->items[i].x;
 				tm.f = text->items[i].y;
 				fz_concat(&trm, &tm, ctm);
-				x = floorf(trm.e);
-				y = floorf(trm.f);
 
-				trunc_trm = trm;
-				trunc_trm.e = QUANT(trm.e - floorf(trm.e), HSUBPIX);
-				trunc_trm.f = QUANT(trm.f - floorf(trm.f), VSUBPIX);
-				scissor.x0 = bbox.x0 - x;
-				scissor.y0 = bbox.y0 - y;
-				scissor.x1 = bbox.x1 - x;
-				scissor.y1 = bbox.y1 - y;
-
-				glyph = fz_render_glyph(dev->ctx, text->font, gid, &trunc_trm, model, scissor);
+				glyph = fz_render_glyph(dev->ctx, text->font, gid, &trm, model, &state->scissor);
 				if (glyph)
 				{
+					int x = (int)trm.e;
+					int y = (int)trm.f;
 					draw_glyph(NULL, mask, glyph, x, y, &bbox);
 					if (state[1].shape)
 						draw_glyph(NULL, state[1].shape, glyph, x, y, &bbox);
-					fz_drop_pixmap(dev->ctx, glyph);
+					fz_drop_glyph(dev->ctx, glyph);
 				}
 				else
 				{
-					fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &trm);
+					fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &tm);
 					if (path)
 					{
 						fz_pixmap *old_dest;
 						float white = 1;
 
-						state = &dev->stack[dev->top];
-						old_dest = state[0].dest;
-						state[0].dest = state[0].mask;
-						state[0].mask = NULL;
+						old_dest = state[1].dest;
+						state[1].dest = state[1].mask;
+						state[1].mask = NULL;
 						fz_try(ctx)
 						{
-							fz_draw_fill_path(devp, path, 0, &fz_identity, fz_device_gray(ctx), &white, 1);
+							fz_draw_fill_path(devp, path, 0, ctm, fz_device_gray(ctx), &white, 1);
 						}
 						fz_always(ctx)
 						{
-							state[0].mask = state[0].dest;
-							state[0].dest = old_dest;
+							state[1].mask = state[1].dest;
+							state[1].dest = old_dest;
 							fz_free_path(dev->ctx, path);
 						}
 						fz_catch(ctx)
@@ -807,13 +835,14 @@ fz_draw_clip_stroke_text(fz_device *devp, fz_text *text, fz_stroke_state *stroke
 	fz_context *ctx = dev->ctx;
 	fz_irect bbox;
 	fz_pixmap *mask, *dest, *shape;
-	fz_matrix tm, trm, trunc_trm;
-	fz_pixmap *glyph;
-	int i, x, y, gid;
+	fz_matrix tm, trm;
+	fz_glyph *glyph;
+	int i, gid;
 	fz_draw_state *state = push_stack(dev);
 	fz_colorspace *model = state->dest->colorspace;
 	fz_rect rect;
 
+	STACK_PUSHED("clip stroke text");
 	/* make the mask the exact size needed */
 	fz_irect_from_rect(&bbox, fz_bound_text(dev->ctx, text, stroke, ctm, &rect));
 	fz_intersect_irect(&bbox, &state->scissor);
@@ -844,7 +873,6 @@ fz_draw_clip_stroke_text(fz_device *devp, fz_text *text, fz_stroke_state *stroke
 
 			for (i = 0; i < text->len; i++)
 			{
-				fz_irect scissor;
 				gid = text->items[i].gid;
 				if (gid < 0)
 					continue;
@@ -852,29 +880,20 @@ fz_draw_clip_stroke_text(fz_device *devp, fz_text *text, fz_stroke_state *stroke
 				tm.e = text->items[i].x;
 				tm.f = text->items[i].y;
 				fz_concat(&trm, &tm, ctm);
-				x = floorf(trm.e);
-				y = floorf(trm.f);
 
-				trunc_trm = trm;
-				trunc_trm.e = QUANT(trm.e - floorf(trm.e), HSUBPIX);
-				trunc_trm.f = QUANT(trm.f - floorf(trm.f), VSUBPIX);
-
-				scissor.x0 = bbox.x0 - x;
-				scissor.y0 = bbox.y0 - y;
-				scissor.x1 = bbox.x1 - x;
-				scissor.y1 = bbox.y1 - y;
-
-				glyph = fz_render_stroked_glyph(dev->ctx, text->font, gid, &trunc_trm, ctm, stroke, scissor);
+				glyph = fz_render_stroked_glyph(dev->ctx, text->font, gid, &trm, ctm, stroke, &state->scissor);
 				if (glyph)
 				{
+					int x = (int)trm.e;
+					int y = (int)trm.f;
 					draw_glyph(NULL, mask, glyph, x, y, &bbox);
 					if (shape)
 						draw_glyph(NULL, shape, glyph, x, y, &bbox);
-					fz_drop_pixmap(dev->ctx, glyph);
+					fz_drop_glyph(dev->ctx, glyph);
 				}
 				else
 				{
-					fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &trm);
+					fz_path *path = fz_outline_glyph(dev->ctx, text->font, gid, &tm);
 					if (path)
 					{
 						fz_pixmap *old_dest;
@@ -886,7 +905,7 @@ fz_draw_clip_stroke_text(fz_device *devp, fz_text *text, fz_stroke_state *stroke
 						state[0].mask = NULL;
 						fz_try(ctx)
 						{
-							fz_draw_stroke_path(devp, path, stroke, &fz_identity, fz_device_gray(ctx), &white, 1);
+							fz_draw_stroke_path(devp, path, stroke, ctm, fz_device_gray(ctx), &white, 1);
 						}
 						fz_always(ctx)
 						{
@@ -1124,7 +1143,7 @@ fz_draw_fill_image(fz_device *devp, fz_image *image, const fz_matrix *ctm, float
 			pixmap = converted;
 		}
 
-		if (dx < pixmap->w && dy < pixmap->h)
+		if (dx < pixmap->w && dy < pixmap->h && !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES))
 		{
 			int gridfit = alpha == 1.0f && !(dev->flags & FZ_DRAWDEV_FLAGS_TYPE3);
 			scaled = fz_transform_pixmap(dev, pixmap, &local_ctm, state->dest->x, state->dest->y, dx, dy, gridfit, &clip);
@@ -1157,7 +1176,7 @@ fz_draw_fill_image(fz_device *devp, fz_image *image, const fz_matrix *ctm, float
 			}
 		}
 
-		fz_paint_image(state->dest, &state->scissor, state->shape, pixmap, &local_ctm, alpha * 255);
+		fz_paint_image(state->dest, &state->scissor, state->shape, pixmap, &local_ctm, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES));
 
 		if (state->blendmode & FZ_BLEND_KNOCKOUT)
 			fz_knockout_end(dev);
@@ -1229,7 +1248,7 @@ fz_draw_fill_image_mask(fz_device *devp, fz_image *image, const fz_matrix *ctm,
 			colorbv[i] = colorfv[i] * 255;
 		colorbv[i] = alpha * 255;
 
-		fz_paint_image_with_color(state->dest, &state->scissor, state->shape, pixmap, &local_ctm, colorbv);
+		fz_paint_image_with_color(state->dest, &state->scissor, state->shape, pixmap, &local_ctm, colorbv, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES));
 
 		if (scaled)
 			fz_drop_pixmap(dev->ctx, scaled);
@@ -1266,6 +1285,7 @@ fz_draw_clip_image_mask(fz_device *devp, fz_image *image, const fz_rect *rect, c
 	fz_matrix local_ctm = *ctm;
 	fz_rect urect;
 
+	STACK_PUSHED("clip image mask");
 	fz_pixmap_bbox(ctx, state->dest, &clip);
 	fz_intersect_irect(&clip, &state->scissor);
 
@@ -1335,7 +1355,7 @@ fz_draw_clip_image_mask(fz_device *devp, fz_image *image, const fz_rect *rect, c
 			if (scaled)
 				pixmap = scaled;
 		}
-		fz_paint_image(mask, &bbox, state->shape, pixmap, &local_ctm, 255);
+		fz_paint_image(mask, &bbox, state->shape, pixmap, &local_ctm, 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES));
 	}
 	fz_always(ctx)
 	{
@@ -1361,6 +1381,7 @@ fz_draw_pop_clip(fz_device *devp)
 		return;
 	}
 	state = &dev->stack[--dev->top];
+	STACK_POPPED("clip");
 
 	/* We can get here with state[1].mask == NULL if the clipping actually
 	 * resolved to a rectangle earlier.
@@ -1383,8 +1404,13 @@ fz_draw_pop_clip(fz_device *devp)
 			fz_paint_pixmap_with_mask(state[0].shape, state[1].shape, state[1].mask);
 			fz_drop_pixmap(dev->ctx, state[1].shape);
 		}
-		fz_drop_pixmap(dev->ctx, state[1].mask);
-		fz_drop_pixmap(dev->ctx, state[1].dest);
+		/* The following tests should not be required, but just occasionally
+		 * errors can cause the stack to get out of sync, and this might save
+		 * our bacon. */
+		if (state[0].mask != state[1].mask)
+			fz_drop_pixmap(dev->ctx, state[1].mask);
+		if (state[0].dest != state[1].dest)
+			fz_drop_pixmap(dev->ctx, state[1].dest);
 #ifdef DUMP_GROUP_BLENDS
 		fz_dump_blend(dev->ctx, state[0].dest, " to get ");
 		if (state[0].shape)
@@ -1410,6 +1436,7 @@ fz_draw_begin_mask(fz_device *devp, const fz_rect *rect, int luminosity, fz_colo
 	fz_pixmap *shape = state->shape;
 	fz_context *ctx = dev->ctx;
 
+	STACK_PUSHED("mask");
 	fz_intersect_irect(fz_irect_from_rect(&bbox, rect), &state->scissor);
 
 	fz_try(ctx)
@@ -1471,41 +1498,48 @@ fz_draw_end_mask(fz_device *devp)
 		return;
 	}
 	state = &dev->stack[dev->top-1];
+	STACK_CONVERT("(mask)");
 	/* pop soft mask buffer */
 	luminosity = state[1].luminosity;
 
 #ifdef DUMP_GROUP_BLENDS
 	dump_spaces(dev->top-1, "Mask -> Clip\n");
 #endif
-	/* convert to alpha mask */
-	temp = fz_alpha_from_gray(dev->ctx, state[1].dest, luminosity);
-	if (state[1].dest != state[0].dest)
-		fz_drop_pixmap(dev->ctx, state[1].dest);
-	state[1].dest = NULL;
-	if (state[1].shape != state[0].shape)
-		fz_drop_pixmap(dev->ctx, state[1].shape);
-	state[1].shape = NULL;
-	if (state[1].mask != state[0].mask)
-		fz_drop_pixmap(dev->ctx, state[1].mask);
-	state[1].mask = NULL;
-
-	/* create new dest scratch buffer */
-	fz_pixmap_bbox(ctx, temp, &bbox);
-	dest = fz_new_pixmap_with_bbox(dev->ctx, state->dest->colorspace, &bbox);
-	fz_clear_pixmap(dev->ctx, dest);
-
-	/* push soft mask as clip mask */
-	state[1].mask = temp;
-	state[1].dest = dest;
-	state[1].blendmode |= FZ_BLEND_ISOLATED;
-	/* If we have a shape, then it'll need to be masked with the
-	 * clip mask when we pop. So create a new shape now. */
-	if (state[0].shape)
+	fz_try(ctx)
 	{
-		state[1].shape = fz_new_pixmap_with_bbox(dev->ctx, NULL, &bbox);
-		fz_clear_pixmap(dev->ctx, state[1].shape);
+		/* convert to alpha mask */
+		temp = fz_alpha_from_gray(dev->ctx, state[1].dest, luminosity);
+		if (state[1].mask != state[0].mask)
+			fz_drop_pixmap(dev->ctx, state[1].mask);
+		state[1].mask = temp;
+		if (state[1].dest != state[0].dest)
+			fz_drop_pixmap(dev->ctx, state[1].dest);
+		state[1].dest = NULL;
+		if (state[1].shape != state[0].shape)
+			fz_drop_pixmap(dev->ctx, state[1].shape);
+		state[1].shape = NULL;
+
+		/* create new dest scratch buffer */
+		fz_pixmap_bbox(ctx, temp, &bbox);
+		dest = fz_new_pixmap_with_bbox(dev->ctx, state->dest->colorspace, &bbox);
+		fz_clear_pixmap(dev->ctx, dest);
+
+		/* push soft mask as clip mask */
+		state[1].dest = dest;
+		state[1].blendmode |= FZ_BLEND_ISOLATED;
+		/* If we have a shape, then it'll need to be masked with the
+		 * clip mask when we pop. So create a new shape now. */
+		if (state[0].shape)
+		{
+			state[1].shape = fz_new_pixmap_with_bbox(dev->ctx, NULL, &bbox);
+			fz_clear_pixmap(dev->ctx, state[1].shape);
+		}
+		state[1].scissor = bbox;
+	}
+	fz_catch(ctx)
+	{
+		emergency_pop_stack(dev, state);
 	}
-	state[1].scissor = bbox;
 }
 
 static void
@@ -1513,7 +1547,7 @@ fz_draw_begin_group(fz_device *devp, const fz_rect *rect, int isolated, int knoc
 {
 	fz_draw_device *dev = devp->user;
 	fz_irect bbox;
-	fz_pixmap *dest, *shape;
+	fz_pixmap *dest;
 	fz_context *ctx = dev->ctx;
 	fz_draw_state *state = &dev->stack[dev->top];
 	fz_colorspace *model = state->dest->colorspace;
@@ -1522,6 +1556,7 @@ fz_draw_begin_group(fz_device *devp, const fz_rect *rect, int isolated, int knoc
 		fz_knockout_begin(dev);
 
 	state = push_stack(dev);
+	STACK_PUSHED("group");
 	fz_intersect_irect(fz_irect_from_rect(&bbox, rect), &state->scissor);
 
 	fz_try(ctx)
@@ -1546,12 +1581,12 @@ fz_draw_begin_group(fz_device *devp, const fz_rect *rect, int isolated, int knoc
 		{
 			/* We can render direct to any existing shape plane.
 			 * If there isn't one, we don't need to make one. */
-			state[1].shape =shape = state[0].shape;
+			state[1].shape = state[0].shape;
 		}
 		else
 		{
-			state[1].shape = shape = fz_new_pixmap_with_bbox(ctx, NULL, &bbox);
-			fz_clear_pixmap(dev->ctx, shape);
+			state[1].shape = fz_new_pixmap_with_bbox(ctx, NULL, &bbox);
+			fz_clear_pixmap(dev->ctx, state[1].shape);
 		}
 
 		state[1].alpha = alpha;
@@ -1585,6 +1620,7 @@ fz_draw_end_group(fz_device *devp)
 	}
 
 	state = &dev->stack[--dev->top];
+	STACK_POPPED("group");
 	alpha = state[1].alpha;
 	blendmode = state[1].blendmode & FZ_BLEND_MODEMASK;
 	isolated = state[1].blendmode & FZ_BLEND_ISOLATED;
@@ -1610,7 +1646,11 @@ fz_draw_end_group(fz_device *devp)
 	else
 		fz_blend_pixmap(state[0].dest, state[1].dest, alpha * 255, blendmode, isolated, state[1].shape);
 
-	fz_drop_pixmap(dev->ctx, state[1].dest);
+	/* The following test should not be required, but just occasionally
+	 * errors can cause the stack to get out of sync, and this might save
+	 * our bacon. */
+	if (state[0].dest != state[1].dest)
+		fz_drop_pixmap(dev->ctx, state[1].dest);
 	if (state[0].shape != state[1].shape)
 	{
 		if (state[0].shape)
@@ -1767,6 +1807,7 @@ fz_draw_begin_tile(fz_device *devp, const fz_rect *area, const fz_rect *view, fl
 		fz_knockout_begin(dev);
 
 	state = push_stack(dev);
+	STACK_PUSHED("tile");
 	fz_irect_from_rect(&bbox, fz_transform_rect(&local_view, ctm));
 	/* We should never have a bbox that entirely covers our destination.
 	 * If we do, then the check for only 1 tile being visible above has
@@ -1859,6 +1900,7 @@ fz_draw_end_tile(fz_device *devp)
 	}
 
 	state = &dev->stack[--dev->top];
+	STACK_PUSHED("tile");
 	xstep = state[1].xstep;
 	ystep = state[1].ystep;
 	area = state[1].area;
@@ -1875,9 +1917,15 @@ fz_draw_end_tile(fz_device *devp)
 	 * single tile) (see fts_15_1506.pdf for an example). This means that
 	 * we have to bias the left hand/bottom edge calculations by the
 	 * difference between the step and the width/height of the tile. */
-	/* state[0].scissor = view, transformed by ctm */
-	x0 = floorf((area.x0 + xstep - state[0].scissor.x1 + state[0].scissor.x0) / xstep);
-	y0 = floorf((area.y0 + ystep - state[0].scissor.y1 + state[0].scissor.y0) / ystep);
+	/* scissor, xstep and area are all in pattern space. */
+	x0 = xstep - scissor.x1 + scissor.x0;
+	if (x0 > 0)
+		x0 = 0;
+	y0 = ystep - scissor.y1 + scissor.y0;
+	if (y0 > 0)
+		y0 = 0;
+	x0 = floorf((area.x0 + x0) / xstep);
+	y0 = floorf((area.y0 + y0) / ystep);
 	x1 = ceilf(area.x1 / xstep);
 	y1 = ceilf(area.y1 / ystep);
 
@@ -1971,8 +2019,13 @@ fz_draw_end_tile(fz_device *devp)
 		/* Do nothing */
 	}
 
-	fz_drop_pixmap(dev->ctx, state[1].dest);
-	fz_drop_pixmap(dev->ctx, state[1].shape);
+	/* The following tests should not be required, but just occasionally
+	 * errors can cause the stack to get out of sync, and this might save
+	 * our bacon. */
+	if (state[0].dest != state[1].dest)
+		fz_drop_pixmap(dev->ctx, state[1].dest);
+	if (state[0].shape != state[1].shape)
+		fz_drop_pixmap(dev->ctx, state[1].shape);
 #ifdef DUMP_GROUP_BLENDS
 	fz_dump_blend(dev->ctx, state[0].dest, " to get ");
 	if (state[0].shape)
@@ -2031,7 +2084,7 @@ fz_new_draw_device(fz_context *ctx, fz_pixmap *dest)
 		ddev->cache_x = fz_new_scale_cache(ctx);
 		ddev->cache_y = fz_new_scale_cache(ctx);
 		ddev->stack = &ddev->init_stack[0];
-		ddev->stack_max = STACK_SIZE;
+		ddev->stack_cap = STACK_SIZE;
 		ddev->stack[0].dest = dest;
 		ddev->stack[0].shape = NULL;
 		ddev->stack[0].mask = NULL;
