@@ -160,7 +160,7 @@ struct fz_edge_s
 struct fz_gel_s
 {
 	fz_rect clip;
-	fz_rect bbox;
+	fz_irect bbox;
 	int cap, len;
 	fz_edge *edges;
 	int acap, alen;
@@ -224,6 +224,7 @@ fz_reset_gel(fz_gel *gel, const fz_irect *clip)
 	gel->bbox.x1 = gel->bbox.y1 = BBOX_MIN;
 
 	gel->len = 0;
+	gel->alen = 0;
 }
 
 void
@@ -254,6 +255,19 @@ fz_bound_gel(const fz_gel *gel, fz_irect *bbox)
 	return bbox;
 }
 
+fz_rect *
+fz_gel_scissor(const fz_gel *gel, fz_rect *r)
+{
+	fz_aa_context *ctxaa = gel->ctx->aa;
+
+	r->x0 = gel->clip.x0 / fz_aa_hscale;
+	r->x1 = gel->clip.x1 / fz_aa_vscale;
+	r->y0 = gel->clip.y0 / fz_aa_hscale;
+	r->y1 = gel->clip.y1 / fz_aa_vscale;
+
+	return r;
+}
+
 enum { INSIDE, OUTSIDE, LEAVE, ENTER };
 
 #define clip_lerp_y(v,m,x0,y0,x1,y1,t) clip_lerp_x(v,m,y0,x0,y1,x1,t)
@@ -312,7 +326,7 @@ fz_insert_gel_raw(fz_gel *gel, int x0, int y0, int x1, int y1)
 	if (y1 > gel->bbox.y1) gel->bbox.y1 = y1;
 
 	if (gel->len + 1 == gel->cap) {
-		int new_cap = gel->cap + 512;
+		int new_cap = gel->cap * 2;
 		gel->edges = fz_resize_array(gel->ctx, gel->edges, new_cap, sizeof(fz_edge));
 		gel->cap = new_cap;
 	}
@@ -413,15 +427,30 @@ fz_insert_gel(fz_gel *gel, float fx0, float fy0, float fx1, float fy1)
 	fz_insert_gel_raw(gel, x0, y0, x1, y1);
 }
 
+static int cmpedge(const void *va, const void *vb)
+{
+	const fz_edge *a = va;
+	const fz_edge *b = vb;
+	return a->y - b->y;
+}
+
 void
 fz_sort_gel(fz_gel *gel)
 {
 	fz_edge *a = gel->edges;
 	int n = gel->len;
-
 	int h, i, k;
 	fz_edge t;
 
+
+	/* quick sort for long lists */
+	if (n > 10000)
+	{
+		qsort(a, n, sizeof *a, cmpedge);
+		return;
+	}
+
+	/* shell sort for short lists */
 	h = 1;
 	if (n < 14) {
 		h = 1;
@@ -445,7 +474,6 @@ fz_sort_gel(fz_gel *gel)
 			}
 			a[k + h] = t;
 		}
-
 		h /= 3;
 	}
 }
@@ -922,7 +950,6 @@ fz_scan_convert_sharp(fz_gel *gel, int eofill, const fz_irect *clip,
 			y += height;
 			if (y >= clip->y0)
 			{
-				height -= y - clip->y0;
 				y = clip->y0;
 				break;
 			}
