@@ -2,6 +2,7 @@
 
 #include <ft2build.h>
 #include FT_FREETYPE_H
+#include FT_ADVANCES_H
 #include FT_STROKER_H
 
 #define MAX_BBOX_TABLE_SIZE 4096
@@ -12,7 +13,7 @@
 static void fz_drop_freetype(fz_context *ctx);
 
 static fz_font *
-fz_new_font(fz_context *ctx, char *name, int use_glyph_bbox, int glyph_count)
+fz_new_font(fz_context *ctx, const char *name, int use_glyph_bbox, int glyph_count)
 {
 	fz_font *font;
 	int i;
@@ -31,9 +32,8 @@ fz_new_font(fz_context *ctx, char *name, int use_glyph_bbox, int glyph_count)
 	font->ft_italic = 0;
 	font->ft_hint = 0;
 
-	font->ft_file = NULL;
-	font->ft_data = NULL;
-	font->ft_size = 0;
+	font->ft_buffer = NULL;
+	font->ft_filepath = NULL;
 
 	font->t3matrix = fz_identity;
 	font->t3resources = NULL;
@@ -151,8 +151,8 @@ fz_drop_font(fz_context *ctx, fz_font *font)
 		fz_drop_freetype(ctx);
 	}
 
-	fz_free(ctx, font->ft_file);
-	fz_free(ctx, font->ft_data);
+	fz_drop_buffer(ctx, font->ft_buffer);
+	fz_free(ctx, font->ft_filepath);
 	fz_free(ctx, font->bbox_table);
 	fz_free(ctx, font->width_table);
 	fz_free(ctx, font);
@@ -187,6 +187,8 @@ struct fz_font_context_s {
 	int ctx_refs;
 	FT_Library ftlib;
 	int ftlib_refs;
+	fz_load_system_font_func load_font;
+	fz_load_system_cjk_font_func load_cjk_font;
 };
 
 #undef __FTERRORS_H__
@@ -206,6 +208,7 @@ void fz_new_font_context(fz_context *ctx)
 	ctx->font->ctx_refs = 1;
 	ctx->font->ftlib = NULL;
 	ctx->font->ftlib_refs = 0;
+	ctx->font->load_font = NULL;
 }
 
 fz_font_context *
@@ -231,6 +234,50 @@ void fz_drop_font_context(fz_context *ctx)
 		fz_free(ctx, ctx->font);
 }
 
+void fz_install_load_system_font_funcs(fz_context *ctx, fz_load_system_font_func f, fz_load_system_cjk_font_func f_cjk)
+{
+	ctx->font->load_font = f;
+	ctx->font->load_cjk_font = f_cjk;
+}
+
+fz_font *fz_load_system_font(fz_context *ctx, const char *name, int bold, int italic, int needs_exact_metrics)
+{
+	fz_font *font = NULL;
+
+	if (ctx->font->load_font)
+	{
+		fz_try(ctx)
+		{
+			font = ctx->font->load_font(ctx, name, bold, italic, needs_exact_metrics);
+		}
+		fz_catch(ctx)
+		{
+			font = NULL;
+		}
+	}
+
+	return font;
+}
+
+fz_font *fz_load_system_cjk_font(fz_context *ctx, const char *name, int ros, int serif)
+{
+	fz_font *font = NULL;
+
+	if (ctx->font->load_cjk_font)
+	{
+		fz_try(ctx)
+		{
+			font = ctx->font->load_cjk_font(ctx, name, ros, serif);
+		}
+		fz_catch(ctx)
+		{
+			font = NULL;
+		}
+	}
+
+	return font;
+}
+
 static const struct ft_error ft_errors[] =
 {
 #include FT_ERRORS_H
@@ -302,7 +349,7 @@ fz_drop_freetype(fz_context *ctx)
 }
 
 fz_font *
-fz_new_font_from_file(fz_context *ctx, char *name, char *path, int index, int use_glyph_bbox)
+fz_new_font_from_file(fz_context *ctx, const char *name, const char *path, int index, int use_glyph_bbox)
 {
 	FT_Face face;
 	fz_font *font;
@@ -329,12 +376,13 @@ fz_new_font_from_file(fz_context *ctx, char *name, char *path, int index, int us
 		(float) face->bbox.yMin / face->units_per_EM,
 		(float) face->bbox.xMax / face->units_per_EM,
 		(float) face->bbox.yMax / face->units_per_EM);
+	font->ft_filepath = fz_strdup(ctx, path);
 
 	return font;
 }
 
 fz_font *
-fz_new_font_from_memory(fz_context *ctx, char *name, unsigned char *data, int len, int index, int use_glyph_bbox)
+fz_new_font_from_memory(fz_context *ctx, const char *name, unsigned char *data, int len, int index, int use_glyph_bbox)
 {
 	FT_Face face;
 	fz_font *font;
@@ -365,6 +413,14 @@ fz_new_font_from_memory(fz_context *ctx, char *name, unsigned char *data, int le
 	return font;
 }
 
+fz_font *
+fz_new_font_from_buffer(fz_context *ctx, const char *name, fz_buffer *buffer, int index, int use_glyph_bbox)
+{
+	fz_font *font = fz_new_font_from_memory(ctx, name, buffer->data, buffer->len, index, use_glyph_bbox);
+	font->ft_buffer = fz_keep_buffer(ctx, buffer); /* remember buffer so we can drop it when we free the font */
+	return font;
+}
+
 static fz_matrix *
 fz_adjust_ft_glyph_width(fz_context *ctx, fz_font *font, int gid, fz_matrix *trm)
 {
@@ -401,58 +457,32 @@ fz_adjust_ft_glyph_width(fz_context *ctx, fz_font *font, int gid, fz_matrix *trm
 	return trm;
 }
 
-static fz_pixmap *
-fz_copy_ft_bitmap(fz_context *ctx, int left, int top, FT_Bitmap *bitmap)
+static fz_glyph *
+glyph_from_ft_bitmap(fz_context *ctx, int left, int top, FT_Bitmap *bitmap)
 {
-	fz_pixmap *pixmap;
-	int y;
-
-	pixmap = fz_new_pixmap(ctx, NULL, bitmap->width, bitmap->rows);
-	pixmap->x = left;
-	pixmap->y = top - bitmap->rows;
-
 	if (bitmap->pixel_mode == FT_PIXEL_MODE_MONO)
-	{
-		for (y = 0; y < pixmap->h; y++)
-		{
-			unsigned char *out = pixmap->samples + (unsigned int)(y * pixmap->w);
-			unsigned char *in = bitmap->buffer + (unsigned int)((pixmap->h - y - 1) * bitmap->pitch);
-			unsigned char bit = 0x80;
-			int w = pixmap->w;
-			while (w--)
-			{
-				*out++ = (*in & bit) ? 255 : 0;
-				bit >>= 1;
-				if (bit == 0)
-				{
-					bit = 0x80;
-					in++;
-				}
-			}
-		}
-	}
+		return fz_new_glyph_from_1bpp_data(ctx, left, top - bitmap->rows, bitmap->width, bitmap->rows, bitmap->buffer + (bitmap->rows-1)*bitmap->pitch, -bitmap->pitch);
 	else
-	{
-		for (y = 0; y < pixmap->h; y++)
-		{
-			memcpy(pixmap->samples + (unsigned int)(y * pixmap->w),
-				bitmap->buffer + (unsigned int)((pixmap->h - y - 1) * bitmap->pitch),
-				pixmap->w);
-		}
-	}
-
-	return pixmap;
+		return fz_new_glyph_from_8bpp_data(ctx, left, top - bitmap->rows, bitmap->width, bitmap->rows, bitmap->buffer + (bitmap->rows-1)*bitmap->pitch, -bitmap->pitch);
 }
 
-/* The glyph cache lock is always taken when this is called. */
-fz_pixmap *
-fz_render_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, int aa)
+static fz_pixmap *
+pixmap_from_ft_bitmap(fz_context *ctx, int left, int top, FT_Bitmap *bitmap)
+{
+	if (bitmap->pixel_mode == FT_PIXEL_MODE_MONO)
+		return fz_new_pixmap_from_1bpp_data(ctx, left, top - bitmap->rows, bitmap->width, bitmap->rows, bitmap->buffer + (bitmap->rows-1)*bitmap->pitch, -bitmap->pitch);
+	else
+		return fz_new_pixmap_from_8bpp_data(ctx, left, top - bitmap->rows, bitmap->width, bitmap->rows, bitmap->buffer + (bitmap->rows-1)*bitmap->pitch, -bitmap->pitch);
+}
+
+/* Takes the freetype lock, and returns with it held */
+static FT_GlyphSlot
+do_ft_render_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, int aa)
 {
 	FT_Face face = font->ft_face;
 	FT_Matrix m;
 	FT_Vector v;
 	FT_Error fterr;
-	fz_pixmap *result;
 	fz_matrix local_trm = *trm;
 
 	float strength = fz_matrix_expansion(trm) * 0.02f;
@@ -488,8 +518,8 @@ fz_render_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm
 		/* enable grid fitting for non-antialiased rendering */
 		float scale = fz_matrix_expansion(&local_trm);
 		m.xx = local_trm.a * 65536 / scale;
-		m.xy = local_trm.b * 65536 / scale;
-		m.yx = local_trm.c * 65536 / scale;
+		m.yx = local_trm.b * 65536 / scale;
+		m.xy = local_trm.c * 65536 / scale;
 		m.yy = local_trm.d * 65536 / scale;
 		v.x = 0;
 		v.y = 0;
@@ -526,7 +556,6 @@ retry_unhinted:
 		if (fterr)
 		{
 			fz_warn(ctx, "freetype load glyph (gid %d): %s", gid, ft_error_string(fterr));
-			fz_unlock(ctx, FZ_LOCK_FREETYPE);
 			return NULL;
 		}
 	}
@@ -541,13 +570,26 @@ retry_unhinted:
 	if (fterr)
 	{
 		fz_warn(ctx, "freetype render glyph (gid %d): %s", gid, ft_error_string(fterr));
+		return NULL;
+	}
+	return face->glyph;
+}
+
+fz_pixmap *
+fz_render_ft_glyph_pixmap(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, int aa)
+{
+	FT_GlyphSlot slot = do_ft_render_glyph(ctx, font, gid, trm, aa);
+	fz_pixmap *pixmap;
+
+	if (slot == NULL)
+	{
 		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
 	fz_try(ctx)
 	{
-		result = fz_copy_ft_bitmap(ctx, face->glyph->bitmap_left, face->glyph->bitmap_top, &face->glyph->bitmap);
+		pixmap = pixmap_from_ft_bitmap(ctx, slot->bitmap_left, slot->bitmap_top, &slot->bitmap);
 	}
 	fz_always(ctx)
 	{
@@ -558,11 +600,41 @@ retry_unhinted:
 		fz_rethrow(ctx);
 	}
 
-	return result;
+	return pixmap;
 }
 
-fz_pixmap *
-fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, const fz_matrix *ctm, fz_stroke_state *state)
+/* The glyph cache lock is always taken when this is called. */
+fz_glyph *
+fz_render_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, int aa)
+{
+	FT_GlyphSlot slot = do_ft_render_glyph(ctx, font, gid, trm, aa);
+	fz_glyph *glyph;
+
+	if (slot == NULL)
+	{
+		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		return NULL;
+	}
+
+	fz_try(ctx)
+	{
+		glyph = glyph_from_ft_bitmap(ctx, slot->bitmap_left, slot->bitmap_top, &slot->bitmap);
+	}
+	fz_always(ctx)
+	{
+		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+
+	return glyph;
+}
+
+/* Takes the freetype lock, and returns with it held */
+static FT_Glyph
+do_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, const fz_matrix *ctm, fz_stroke_state *state)
 {
 	FT_Face face = font->ft_face;
 	float expansion = fz_matrix_expansion(ctm);
@@ -572,8 +644,6 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	FT_Error fterr;
 	FT_Stroker stroker;
 	FT_Glyph glyph;
-	FT_BitmapGlyph bitmap;
-	fz_pixmap *pixmap;
 	FT_Stroker_LineJoin line_join;
 	fz_matrix local_trm = *trm;
 
@@ -594,7 +664,6 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	if (fterr)
 	{
 		fz_warn(ctx, "FT_Set_Char_Size: %s", ft_error_string(fterr));
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
@@ -604,7 +673,6 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	if (fterr)
 	{
 		fz_warn(ctx, "FT_Load_Glyph(gid %d): %s", gid, ft_error_string(fterr));
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
@@ -612,7 +680,6 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	if (fterr)
 	{
 		fz_warn(ctx, "FT_Stroker_New: %s", ft_error_string(fterr));
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
@@ -638,7 +705,6 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	{
 		fz_warn(ctx, "FT_Get_Glyph: %s", ft_error_string(fterr));
 		FT_Stroker_Done(stroker);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
@@ -648,7 +714,6 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 		fz_warn(ctx, "FT_Glyph_Stroke: %s", ft_error_string(fterr));
 		FT_Done_Glyph(glyph);
 		FT_Stroker_Done(stroker);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
@@ -659,14 +724,27 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	{
 		fz_warn(ctx, "FT_Glyph_To_Bitmap: %s", ft_error_string(fterr));
 		FT_Done_Glyph(glyph);
+		return NULL;
+	}
+	return glyph;
+}
+
+fz_pixmap *
+fz_render_ft_stroked_glyph_pixmap(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, const fz_matrix *ctm, fz_stroke_state *state)
+{
+	FT_Glyph glyph = do_render_ft_stroked_glyph(ctx, font, gid, trm, ctm, state);
+	FT_BitmapGlyph bitmap = (FT_BitmapGlyph)glyph;
+	fz_pixmap *pixmap;
+
+	if (bitmap == NULL)
+	{
 		fz_unlock(ctx, FZ_LOCK_FREETYPE);
 		return NULL;
 	}
 
-	bitmap = (FT_BitmapGlyph)glyph;
 	fz_try(ctx)
 	{
-		pixmap = fz_copy_ft_bitmap(ctx, bitmap->left, bitmap->top, &bitmap->bitmap);
+		pixmap = pixmap_from_ft_bitmap(ctx, bitmap->left, bitmap->top, &bitmap->bitmap);
 	}
 	fz_always(ctx)
 	{
@@ -681,40 +759,85 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_mat
 	return pixmap;
 }
 
+fz_glyph *
+fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, const fz_matrix *ctm, fz_stroke_state *state)
+{
+	FT_Glyph glyph = do_render_ft_stroked_glyph(ctx, font, gid, trm, ctm, state);
+	FT_BitmapGlyph bitmap = (FT_BitmapGlyph)glyph;
+	fz_glyph *result;
+
+	if (bitmap == NULL)
+	{
+		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		return NULL;
+	}
+
+	fz_try(ctx)
+	{
+		result = glyph_from_ft_bitmap(ctx, bitmap->left, bitmap->top, &bitmap->bitmap);
+	}
+	fz_always(ctx)
+	{
+		FT_Done_Glyph(glyph);
+		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+
+	return result;
+}
+
 static fz_rect *
-fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, fz_rect *bounds)
+fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid, fz_rect *bounds)
 {
 	FT_Face face = font->ft_face;
 	FT_Error fterr;
 	FT_BBox cbox;
 	FT_Matrix m;
 	FT_Vector v;
+	int ft_flags;
 
 	// TODO: refactor loading into fz_load_ft_glyph
 	// TODO: cache results
 
-	float strength = fz_matrix_expansion(trm) * 0.02f;
-	fz_matrix local_trm = *trm;
+	const int scale = face->units_per_EM;
+	const float recip = 1 / (float)scale;
+	const float strength = 0.02f;
+	fz_matrix local_trm = fz_identity;
 
 	fz_adjust_ft_glyph_width(ctx, font, gid, &local_trm);
 
 	if (font->ft_italic)
 		fz_pre_shear(&local_trm, SHEAR, 0);
 
-	m.xx = local_trm.a * 64; /* should be 65536 */
-	m.yx = local_trm.b * 64;
-	m.xy = local_trm.c * 64;
-	m.yy = local_trm.d * 64;
-	v.x = local_trm.e * 64;
-	v.y = local_trm.f * 64;
+	m.xx = local_trm.a * 65536;
+	m.yx = local_trm.b * 65536;
+	m.xy = local_trm.c * 65536;
+	m.yy = local_trm.d * 65536;
+	v.x = local_trm.e * 65536;
+	v.y = local_trm.f * 65536;
+
+	if (font->ft_hint)
+	{
+		ft_flags = FT_LOAD_NO_BITMAP;
+	}
+	else
+	{
+		ft_flags = FT_LOAD_NO_BITMAP | FT_LOAD_NO_HINTING;
+	}
 
 	fz_lock(ctx, FZ_LOCK_FREETYPE);
-	fterr = FT_Set_Char_Size(face, 65536, 65536, 72, 72); /* should be 64, 64 */
+	/* Set the char size to scale=face->units_per_EM to effectively give
+	 * us unscaled results. This avoids quantisation. We then apply the
+	 * scale ourselves below. */
+	fterr = FT_Set_Char_Size(face, scale, scale, 72, 72);
 	if (fterr)
 		fz_warn(ctx, "freetype setting character size: %s", ft_error_string(fterr));
 	FT_Set_Transform(face, &m, &v);
 
-	fterr = FT_Load_Glyph(face, gid, FT_LOAD_NO_BITMAP | FT_LOAD_NO_HINTING);
+	fterr = FT_Load_Glyph(face, gid, ft_flags);
 	if (fterr)
 	{
 		fz_warn(ctx, "freetype load glyph (gid %d): %s", gid, ft_error_string(fterr));
@@ -726,16 +849,16 @@ fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm,
 
 	if (font->ft_bold)
 	{
-		FT_Outline_Embolden(&face->glyph->outline, strength * 64);
-		FT_Outline_Translate(&face->glyph->outline, -strength * 32, -strength * 32);
+		FT_Outline_Embolden(&face->glyph->outline, strength * scale);
+		FT_Outline_Translate(&face->glyph->outline, -strength * 0.5 * scale, -strength * 0.5 * scale);
 	}
 
 	FT_Outline_Get_CBox(&face->glyph->outline, &cbox);
 	fz_unlock(ctx, FZ_LOCK_FREETYPE);
-	bounds->x0 = cbox.xMin / 64.0f;
-	bounds->y0 = cbox.yMin / 64.0f;
-	bounds->x1 = cbox.xMax / 64.0f;
-	bounds->y1 = cbox.yMax / 64.0f;
+	bounds->x0 = cbox.xMin * recip;
+	bounds->y0 = cbox.yMin * recip;
+	bounds->x1 = cbox.xMax * recip;
+	bounds->y1 = cbox.yMax * recip;
 
 	if (fz_is_empty_rect(bounds))
 	{
@@ -751,57 +874,66 @@ fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm,
 struct closure {
 	fz_context *ctx;
 	fz_path *path;
-	float x, y;
+	fz_matrix trm;
 };
 
-static int move_to(const FT_Vector *p, void *cc)
+static int move_to(const FT_Vector *p, void *cc_)
 {
-	fz_context *ctx = ((struct closure *)cc)->ctx;
-	fz_path *path = ((struct closure *)cc)->path;
-	float tx = ((struct closure *)cc)->x;
-	float ty = ((struct closure *)cc)->y;
-	fz_moveto(ctx, path, tx + p->x / 64.0f, ty + p->y / 64.0f);
+	struct closure *cc = (struct closure *)cc_;
+	fz_context *ctx = cc->ctx;
+	fz_path *path = cc->path;
+	fz_point pt;
+
+	fz_transform_point_xy(&pt, &cc->trm, p->x, p->y);
+	fz_moveto(ctx, path, pt.x, pt.y);
 	return 0;
 }
 
-static int line_to(const FT_Vector *p, void *cc)
+static int line_to(const FT_Vector *p, void *cc_)
 {
-	fz_context *ctx = ((struct closure *)cc)->ctx;
-	fz_path *path = ((struct closure *)cc)->path;
-	float tx = ((struct closure *)cc)->x;
-	float ty = ((struct closure *)cc)->y;
-	fz_lineto(ctx, path, tx + p->x / 64.0f, ty + p->y / 64.0f);
+	struct closure *cc = (struct closure *)cc_;
+	fz_context *ctx = cc->ctx;
+	fz_path *path = cc->path;
+	fz_point pt;
+
+	fz_transform_point_xy(&pt, &cc->trm, p->x, p->y);
+	fz_lineto(ctx, path, pt.x, pt.y);
 	return 0;
 }
 
-static int conic_to(const FT_Vector *c, const FT_Vector *p, void *cc)
+static int conic_to(const FT_Vector *c, const FT_Vector *p, void *cc_)
 {
-	fz_context *ctx = ((struct closure *)cc)->ctx;
-	fz_path *path = ((struct closure *)cc)->path;
-	float tx = ((struct closure *)cc)->x;
-	float ty = ((struct closure *)cc)->y;
+	struct closure *cc = (struct closure *)cc_;
+	fz_context *ctx = cc->ctx;
+	fz_path *path = cc->path;
+	fz_point ct, pt;
 	fz_point s, c1, c2;
-	float cx = tx + c->x / 64.0f, cy = ty + c->y / 64.0f;
-	float px = tx + p->x / 64.0f, py = ty + p->y / 64.0f;
+
+	fz_transform_point_xy(&ct, &cc->trm, c->x, c->y);
+	fz_transform_point_xy(&pt, &cc->trm, p->x, p->y);
+
 	s = fz_currentpoint(ctx, path);
-	c1.x = (s.x + cx * 2) / 3;
-	c1.y = (s.y + cy * 2) / 3;
-	c2.x = (px + cx * 2) / 3;
-	c2.y = (py + cy * 2) / 3;
-	fz_curveto(ctx, path, c1.x, c1.y, c2.x, c2.y, px, py);
+	c1.x = (s.x + ct.x * 2) / 3;
+	c1.y = (s.y + ct.y * 2) / 3;
+	c2.x = (pt.x + ct.x * 2) / 3;
+	c2.y = (pt.y + ct.y * 2) / 3;
+
+	fz_curveto(ctx, path, c1.x, c1.y, c2.x, c2.y, pt.x, pt.y);
 	return 0;
 }
 
-static int cubic_to(const FT_Vector *c1, const FT_Vector *c2, const FT_Vector *p, void *cc)
+static int cubic_to(const FT_Vector *c1, const FT_Vector *c2, const FT_Vector *p, void *cc_)
 {
-	fz_context *ctx = ((struct closure *)cc)->ctx;
-	fz_path *path = ((struct closure *)cc)->path;
-	float tx = ((struct closure *)cc)->x;
-	float ty = ((struct closure *)cc)->y;
-	fz_curveto(ctx, path,
-		tx + c1->x/64.0f, ty + c1->y/64.0f,
-		tx + c2->x/64.0f, ty + c2->y/64.0f,
-		tx + p->x/64.0f, ty + p->y/64.0f);
+	struct closure *cc = (struct closure *)cc_;
+	fz_context *ctx = cc->ctx;
+	fz_path *path = cc->path;
+	fz_point c1t, c2t, pt;
+
+	fz_transform_point_xy(&c1t, &cc->trm, c1->x, c1->y);
+	fz_transform_point_xy(&c2t, &cc->trm, c2->x, c2->y);
+	fz_transform_point_xy(&pt, &cc->trm, p->x, p->y);
+
+	fz_curveto(ctx, path, c1t.x, c1t.y, c2t.x, c2t.y, pt.x, pt.y);
 	return 0;
 }
 
@@ -814,33 +946,34 @@ fz_outline_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *tr
 {
 	struct closure cc;
 	FT_Face face = font->ft_face;
-	FT_Matrix m;
-	FT_Vector v;
 	int fterr;
 	fz_matrix local_trm = *trm;
+	int ft_flags;
 
-	float strength = fz_matrix_expansion(trm) * 0.02f;
+	const int scale = face->units_per_EM;
+	const float recip = 1 / (float)scale;
+	const float strength = 0.02f;
 
 	fz_adjust_ft_glyph_width(ctx, font, gid, &local_trm);
 
 	if (font->ft_italic)
 		fz_pre_shear(&local_trm, SHEAR, 0);
 
-	m.xx = local_trm.a * 64; /* should be 65536 */
-	m.yx = local_trm.b * 64;
-	m.xy = local_trm.c * 64;
-	m.yy = local_trm.d * 64;
-	v.x = 0;
-	v.y = 0;
-
 	fz_lock(ctx, FZ_LOCK_FREETYPE);
 
-	fterr = FT_Set_Char_Size(face, 65536, 65536, 72, 72); /* should be 64, 64 */
-	if (fterr)
-		fz_warn(ctx, "freetype setting character size: %s", ft_error_string(fterr));
-	FT_Set_Transform(face, &m, &v);
+	if (font->ft_hint)
+	{
+		ft_flags = FT_LOAD_NO_BITMAP | FT_LOAD_IGNORE_TRANSFORM;
+		fterr = FT_Set_Char_Size(face, scale, scale, 72, 72);
+		if (fterr)
+			fz_warn(ctx, "freetype setting character size: %s", ft_error_string(fterr));
+	}
+	else
+	{
+		ft_flags = FT_LOAD_NO_SCALE | FT_LOAD_IGNORE_TRANSFORM;
+	}
 
-	fterr = FT_Load_Glyph(face, gid, FT_LOAD_NO_BITMAP | FT_LOAD_NO_HINTING);
+	fterr = FT_Load_Glyph(face, gid, ft_flags);
 	if (fterr)
 	{
 		fz_warn(ctx, "freetype load glyph (gid %d): %s", gid, ft_error_string(fterr));
@@ -850,17 +983,17 @@ fz_outline_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *tr
 
 	if (font->ft_bold)
 	{
-		FT_Outline_Embolden(&face->glyph->outline, strength * 64);
-		FT_Outline_Translate(&face->glyph->outline, -strength * 32, -strength * 32);
+		FT_Outline_Embolden(&face->glyph->outline, strength * scale);
+		FT_Outline_Translate(&face->glyph->outline, -strength * 0.5 * scale, -strength * 0.5 * scale);
 	}
 
+	cc.path = NULL;
 	fz_try(ctx)
 	{
 		cc.ctx = ctx;
 		cc.path = fz_new_path(ctx);
-		cc.x = local_trm.e;
-		cc.y = local_trm.f;
-		fz_moveto(ctx, cc.path, cc.x, cc.y);
+		fz_concat(&cc.trm, fz_scale(&cc.trm, recip, recip), &local_trm);
+		fz_moveto(ctx, cc.path, cc.trm.e, cc.trm.f);
 		FT_Outline_Decompose(&face->glyph->outline, &outline_funcs, &cc);
 		fz_closepath(ctx, cc.path);
 	}
@@ -883,7 +1016,7 @@ fz_outline_ft_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *tr
  */
 
 fz_font *
-fz_new_type3_font(fz_context *ctx, char *name, const fz_matrix *matrix)
+fz_new_type3_font(fz_context *ctx, const char *name, const fz_matrix *matrix)
 {
 	fz_font *font;
 	int i;
@@ -941,6 +1074,7 @@ fz_bound_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm,
 	fz_display_list *list;
 	fz_matrix ctm;
 	fz_device *dev;
+	fz_rect big;
 
 	list = font->t3lists[gid];
 	if (!list)
@@ -964,14 +1098,32 @@ fz_bound_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm,
 		fz_rethrow(ctx);
 	}
 
+	/* clip the bbox size to a reasonable maximum for degenerate glyphs */
+	big = font->bbox;
+	fz_expand_rect(&big, fz_max(fz_matrix_expansion(&ctm) * 2, fz_max(big.x1 - big.x0, big.y1 - big.y0)));
+	fz_intersect_rect(bounds, &big);
+
 	return bounds;
 }
 
-fz_pixmap *
-fz_render_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, fz_colorspace *model, fz_irect scissor)
+void
+fz_run_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, fz_device *dev)
 {
 	fz_display_list *list;
 	fz_matrix ctm;
+
+	list = font->t3lists[gid];
+	if (!list)
+		return;
+
+	fz_concat(&ctm, &font->t3matrix, trm);
+	fz_run_display_list(list, dev, &ctm, &fz_infinite_rect, NULL);
+}
+
+fz_pixmap *
+fz_render_t3_glyph_pixmap(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, fz_colorspace *model, const fz_irect *scissor)
+{
+	fz_display_list *list;
 	fz_rect bounds;
 	fz_irect bbox;
 	fz_device *dev;
@@ -1004,20 +1156,39 @@ fz_render_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm
 
 	fz_expand_rect(fz_bound_glyph(ctx, font, gid, trm, &bounds), 1);
 	fz_irect_from_rect(&bbox, &bounds);
-	fz_intersect_irect(&bbox, &scissor);
+	fz_intersect_irect(&bbox, scissor);
 
 	glyph = fz_new_pixmap_with_bbox(ctx, model ? model : fz_device_gray(ctx), &bbox);
 	fz_clear_pixmap(ctx, glyph);
 
-	fz_concat(&ctm, &font->t3matrix, trm);
 	dev = fz_new_draw_device_type3(ctx, glyph);
-	fz_run_display_list(list, dev, &ctm, &fz_infinite_rect, NULL);
-	fz_free_device(dev);
+	fz_try(ctx)
+	{
+		fz_run_t3_glyph(ctx, font, gid, trm, dev);
+	}
+	fz_always(ctx)
+	{
+		fz_free_device(dev);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
 
 	if (!model)
 	{
-		result = fz_alpha_from_gray(ctx, glyph, 0);
-		fz_drop_pixmap(ctx, glyph);
+		fz_try(ctx)
+		{
+			result = fz_alpha_from_gray(ctx, glyph, 0);
+		}
+		fz_always(ctx)
+		{
+			fz_drop_pixmap(ctx, glyph);
+		}
+		fz_catch(ctx)
+		{
+			fz_rethrow(ctx);
+		}
 	}
 	else
 		result = glyph;
@@ -1025,6 +1196,13 @@ fz_render_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm
 	return result;
 }
 
+fz_glyph *
+fz_render_t3_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, fz_colorspace *model, const fz_irect *scissor)
+{
+	fz_pixmap *pixmap = fz_render_t3_glyph_pixmap(ctx, font, gid, trm, model, scissor);
+	return fz_new_glyph_from_pixmap(ctx, pixmap);
+}
+
 void
 fz_render_t3_glyph_direct(fz_context *ctx, fz_device *dev, fz_font *font, int gid, const fz_matrix *trm, void *gstate, int nested_depth)
 {
@@ -1091,7 +1269,7 @@ fz_bound_glyph(fz_context *ctx, fz_font *font, int gid, const fz_matrix *trm, fz
 		if (fz_is_infinite_rect(&font->bbox_table[gid]))
 		{
 			if (font->ft_face)
-				fz_bound_ft_glyph(ctx, font, gid, &fz_identity, &font->bbox_table[gid]);
+				fz_bound_ft_glyph(ctx, font, gid, &font->bbox_table[gid]);
 			else if (font->t3lists)
 				fz_bound_t3_glyph(ctx, font, gid, &fz_identity, &font->bbox_table[gid]);
 			else
@@ -1122,3 +1300,48 @@ int fz_glyph_cacheable(fz_context *ctx, fz_font *font, int gid)
 		return 1;
 	return (font->t3flags[gid] & FZ_DEVFLAG_UNCACHEABLE) == 0;
 }
+
+static float
+fz_advance_ft_glyph(fz_context *ctx, fz_font *font, int gid)
+{
+	FT_Fixed adv;
+	int mask = FT_LOAD_NO_SCALE | FT_LOAD_IGNORE_TRANSFORM;
+
+	if (font->ft_substitute && font->width_table && gid < font->width_count)
+		return font->width_table[gid];
+
+	FT_Get_Advance(font->ft_face, gid, mask, &adv);
+	return (float) adv / ((FT_Face)font->ft_face)->units_per_EM;
+}
+
+static float
+fz_advance_t3_glyph(fz_context *ctx, fz_font *font, int gid)
+{
+	if (gid < 0 || gid > 255)
+		return 0;
+	return font->t3widths[gid];
+}
+
+float
+fz_advance_glyph(fz_context *ctx, fz_font *font, int gid)
+{
+	if (font->ft_face)
+		return fz_advance_ft_glyph(ctx, font, gid);
+	if (font->t3procs)
+		return fz_advance_t3_glyph(ctx, font, gid);
+	return 0;
+}
+
+static int
+fz_encode_ft_character(fz_context *ctx, fz_font *font, int ucs)
+{
+	return FT_Get_Char_Index(font->ft_face, ucs);
+}
+
+int
+fz_encode_character(fz_context *ctx, fz_font *font, int ucs)
+{
+	if (font->ft_face)
+		return fz_encode_ft_character(ctx, font, ucs);
+	return ucs;
+}
