@@ -233,8 +233,9 @@ pdf_load_substitute_cjk_font(fz_context *ctx, pdf_font_desc *fontdesc, char *fon
 		if (!data)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find builtin CJK font");
 
-		/* a glyph bbox cache is too big for droid sans fallback (51k glyphs!) */
-		fontdesc->font = fz_new_font_from_memory(ctx, fontname, data, len, 0, 0);
+		/* The DroidSansFallback fonts have both H and V variants in the TTC */
+		/* A glyph bbox cache is too big for DroidSansFallback (51k glyphs!) */
+		fontdesc->font = fz_new_font_from_memory(ctx, fontname, data, len, fontdesc->wmode, 0);
 	}
 
 	fontdesc->font->ft_substitute = 1;
@@ -848,11 +849,12 @@ load_cid_font(pdf_document *doc, pdf_obj *dict, pdf_obj *encoding, pdf_obj *to_u
 	pdf_obj *widths;
 	pdf_obj *descriptor;
 	pdf_font_desc *fontdesc = NULL;
+	pdf_cmap *cmap;
 	FT_Face face;
-	int kind;
 	char collection[256];
 	char *basefont;
 	int i, k, fterr;
+	pdf_obj *cidtogidmap;
 	pdf_obj *obj;
 	int dw;
 	fz_context *ctx = doc->ctx;
@@ -889,83 +891,81 @@ load_cid_font(pdf_document *doc, pdf_obj *dict, pdf_obj *encoding, pdf_obj *to_u
 			fz_strlcat(collection, tmpstr, sizeof collection);
 		}
 
+		/* Encoding */
+
+		if (pdf_is_name(encoding))
+		{
+			if (!strcmp(pdf_to_name(encoding), "Identity-H"))
+				cmap = pdf_new_identity_cmap(ctx, 0, 2);
+			else if (!strcmp(pdf_to_name(encoding), "Identity-V"))
+				cmap = pdf_new_identity_cmap(ctx, 1, 2);
+			else
+				cmap = pdf_load_system_cmap(ctx, pdf_to_name(encoding));
+		}
+		else if (pdf_is_indirect(encoding))
+		{
+			cmap = pdf_load_embedded_cmap(doc, encoding);
+		}
+		else
+		{
+			fz_throw(ctx, FZ_ERROR_GENERIC, "syntaxerror: font missing encoding");
+		}
+
 		/* Load font file */
 
 		fontdesc = pdf_new_font_desc(ctx);
 
+		fontdesc->encoding = cmap;
+		fontdesc->size += pdf_cmap_size(ctx, fontdesc->encoding);
+
+		pdf_set_font_wmode(ctx, fontdesc, pdf_cmap_wmode(ctx, fontdesc->encoding));
+
 		descriptor = pdf_dict_gets(dict, "FontDescriptor");
 		if (!descriptor)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "syntaxerror: missing font descriptor");
 		pdf_load_font_descriptor(fontdesc, doc, descriptor, collection, basefont, 1);
 
 		face = fontdesc->font->ft_face;
-		kind = ft_kind(face);
 
-		/* Encoding */
+		/* Apply encoding */
 
-		if (pdf_is_name(encoding))
+		cidtogidmap = pdf_dict_gets(dict, "CIDToGIDMap");
+		if (pdf_is_indirect(cidtogidmap))
 		{
-			if (!strcmp(pdf_to_name(encoding), "Identity-H"))
-				fontdesc->encoding = pdf_new_identity_cmap(ctx, 0, 2);
-			else if (!strcmp(pdf_to_name(encoding), "Identity-V"))
-				fontdesc->encoding = pdf_new_identity_cmap(ctx, 1, 2);
-			else
-				fontdesc->encoding = pdf_load_system_cmap(ctx, pdf_to_name(encoding));
+			fz_buffer *buf;
+
+			buf = pdf_load_stream(doc, pdf_to_num(cidtogidmap), pdf_to_gen(cidtogidmap));
+
+			fontdesc->cid_to_gid_len = (buf->len) / 2;
+			fontdesc->cid_to_gid = fz_malloc_array(ctx, fontdesc->cid_to_gid_len, sizeof(unsigned short));
+			fontdesc->size += fontdesc->cid_to_gid_len * sizeof(unsigned short);
+			for (i = 0; i < fontdesc->cid_to_gid_len; i++)
+				fontdesc->cid_to_gid[i] = (buf->data[i * 2] << 8) + buf->data[i * 2 + 1];
+
+			fz_drop_buffer(ctx, buf);
 		}
-		else if (pdf_is_indirect(encoding))
-		{
-			fontdesc->encoding = pdf_load_embedded_cmap(doc, encoding);
-		}
-		else
-		{
-			fz_throw(ctx, FZ_ERROR_GENERIC, "syntaxerror: font missing encoding");
-		}
-		fontdesc->size += pdf_cmap_size(ctx, fontdesc->encoding);
 
-		pdf_set_font_wmode(ctx, fontdesc, pdf_cmap_wmode(ctx, fontdesc->encoding));
-
-		if (kind == TRUETYPE)
+		/* if font is external, cidtogidmap should not be identity */
+		/* so we map from cid to unicode and then map that through the (3 1) */
+		/* unicode cmap to get a glyph id */
+		else if (fontdesc->font->ft_substitute)
 		{
-			pdf_obj *cidtogidmap;
-
-			cidtogidmap = pdf_dict_gets(dict, "CIDToGIDMap");
-			if (pdf_is_indirect(cidtogidmap))
+			fterr = FT_Select_Charmap(face, ft_encoding_unicode);
+			if (fterr)
 			{
-				fz_buffer *buf;
-
-				buf = pdf_load_stream(doc, pdf_to_num(cidtogidmap), pdf_to_gen(cidtogidmap));
-
-				fontdesc->cid_to_gid_len = (buf->len) / 2;
-				fontdesc->cid_to_gid = fz_malloc_array(ctx, fontdesc->cid_to_gid_len, sizeof(unsigned short));
-				fontdesc->size += fontdesc->cid_to_gid_len * sizeof(unsigned short);
-				for (i = 0; i < fontdesc->cid_to_gid_len; i++)
-					fontdesc->cid_to_gid[i] = (buf->data[i * 2] << 8) + buf->data[i * 2 + 1];
-
-				fz_drop_buffer(ctx, buf);
+				fz_throw(ctx, FZ_ERROR_GENERIC, "fonterror: no unicode cmap when emulating CID font: %s", ft_error_string(fterr));
 			}
 
-			/* if truetype font is external, cidtogidmap should not be identity */
-			/* so we map from cid to unicode and then map that through the (3 1) */
-			/* unicode cmap to get a glyph id */
-			else if (fontdesc->font->ft_substitute)
-			{
-				fterr = FT_Select_Charmap(face, ft_encoding_unicode);
-				if (fterr)
-				{
-					fz_throw(ctx, FZ_ERROR_GENERIC, "fonterror: no unicode cmap when emulating CID font: %s", ft_error_string(fterr));
-				}
-
-				if (!strcmp(collection, "Adobe-CNS1"))
-					fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-CNS1-UCS2");
-				else if (!strcmp(collection, "Adobe-GB1"))
-					fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-GB1-UCS2");
-				else if (!strcmp(collection, "Adobe-Japan1"))
-					fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-Japan1-UCS2");
-				else if (!strcmp(collection, "Adobe-Japan2"))
-					fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-Japan2-UCS2");
-				else if (!strcmp(collection, "Adobe-Korea1"))
-					fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-Korea1-UCS2");
-			}
+			if (!strcmp(collection, "Adobe-CNS1"))
+				fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-CNS1-UCS2");
+			else if (!strcmp(collection, "Adobe-GB1"))
+				fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-GB1-UCS2");
+			else if (!strcmp(collection, "Adobe-Japan1"))
+				fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-Japan1-UCS2");
+			else if (!strcmp(collection, "Adobe-Japan2"))
+				fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-Japan2-UCS2");
+			else if (!strcmp(collection, "Adobe-Korea1"))
+				fontdesc->to_ttf_cmap = pdf_load_system_cmap(ctx, "Adobe-Korea1-UCS2");
 		}
 
 		pdf_load_to_unicode(doc, fontdesc, NULL, collection, to_unicode);
