@@ -31,9 +31,9 @@ pdf_load_outline_imp(pdf_document *doc, pdf_obj *dict)
 			if (obj)
 				node->title = pdf_to_utf8(doc, obj);
 
-			if ((obj = pdf_dict_gets(dict, "Dest")))
-				node->dest = pdf_parse_link_dest(doc, obj);
-			else if ((obj = pdf_dict_gets(dict, "A")))
+			if ((obj = pdf_dict_gets(dict, "Dest")) != NULL)
+				node->dest = pdf_parse_link_dest(doc, FZ_LINK_GOTO, obj);
+			else if ((obj = pdf_dict_gets(dict, "A")) != NULL)
 				node->dest = pdf_parse_action(doc, obj);
 
 			obj = pdf_dict_gets(dict, "First");
